
# coding: utf-8

# # Computational design of growth-coupled production strains
# Specialisation course at NNF Center for Biosustainability <br>
# Student: Valentijn Broeken (s172092) <br>
# Daily supervisor: Kristian Jensen <br>
# Spring semester 2018
# 
# The purpose of the algorithm is to propose strategies that growth-couple production through a target reaction, using metabolic modelling. The knock-outs, knock-ins and medium additions suggested make that if the cell grows at its maximal growth rate, a flux through the target reaction is established – i.e. weak or holistic growth coupling is realised in the model. The algorithm presented consists of a mixed integer linear programming (MILP) problem that maximizes the cell’s growth coupling potential: the difference in the optimal growth rate when the target reaction is present or absent. In other words: the MILP finds designs that maximally increase the optimal growth rate when the target reaction is present, compared to the optimal growth rate without the target reaction. 
# 
# **This notebook contains the final version of the growth-coupling algorithm as created in this specialisation course.**

# In[1]:


# Loading dependencies
import os
os.environ["OPTLANG_USE_SYMENGINE"] = "False"
import cameo
import cobra
from cobra.test import create_test_model
from optlang.duality import convert_linear_problem_to_dual
import logging
from cameo.models import universal
from cameo.data import metanetx

logger = logging.getLogger()

cobra.show_versions()


# # Ecolicore model
# The EColiCore model by Orth, Flemming and Palsson (2011), consisting of 72 metabolites and 95 reactions. It is very small, so mostly useful for testing and debugging.

# In[70]:


# Load EColiCore model (2010), consisting of 74 metabolites and 95 reactions
model = cameo.load_model("e_coli_core")
# Set up: solver, make copy of the model, remove ATP maintenance reaction
model.solver = "cplex"

print(model.optimize())
model.remove_reactions([model.reactions.ATPM])
print(model.optimize())
model


# In[ ]:


# Add target reaction cadA + the metabolite and the exporter for itaconic acid
ITA_ADDED = cobra.Reaction("ITA_ADDED", name="cis-Aconitate decarboxylase")
EX_ITA = cobra.Reaction("EX_ITA", name="Itaconic acid exporter")
model.add_reactions([ITA_ADDED, EX_ITA])
model.reactions.ITA_ADDED.reaction = "acon_C_c + h_c -> itacon_c + co2_c"
model.reactions.EX_ITA.reaction = "itacon_c -->"
model.reactions.ITA_ADDED.gene_reaction_rule = "cadA"
model.metabolites.itacon_c.name = "Itaconic acid"
model.metabolites.itacon_c.formula = "C5H4O4"
model.metabolites.itacon_c.charge = -2
model.metabolites.itacon_c.compartment = "c"

orig_model = model.copy()


# # EColiCore2 Model
# The EColiCore2 model by Hädicke and Klamt (2017), consisting of 486 metabolites and 499 reactions. 
# 
# As the model is not available in BiGG standard from cameo, it was derived again from the iJO1366 and saved as a json-file. The supplementary file by Hädicke and Klamt (2017) does namely not contain all reaction and metabolite information that is useful to have when exploring the solutions. The json file is created as in the "03 TheAlgorithm EcoliCore2 - methylation" notebook, according to the BiGG standard of the iJO1366 model.

# In[ ]:


model = cobra.io.load_json_model("models/EColiCore2_from_iJO1366.json")


# In[ ]:


model.solver = "cplex"
model.objective = "BIOMASS_Ec_iJO1366_core_53p95M"
model.objective.direction = "max"
print(model.optimize())
model.remove_reactions([model.reactions.ATPM])
print(model.optimize())


# In[ ]:


# For the itaconic acid case:
# Add target reaction cadA + the metabolite and the exporter
ITA_ADDED = cobra.Reaction("ITA_ADDED", name="cis-Aconitate decarboxylase")
EX_ITA = cobra.Reaction("EX_ITA", name="Itaconic acid exporter")
model.add_reactions([ITA_ADDED, EX_ITA])
model.reactions.ITA_ADDED.reaction = "acon_C_c + h_c -> itacon_c + co2_c"
model.reactions.EX_ITA.reaction = "itacon_c -->"
model.reactions.ITA_ADDED.gene_reaction_rule = "cadA"
model.metabolites.itacon_c.name = "Itaconic acid"
model.metabolites.itacon_c.formula = "C5H4O4"
model.metabolites.itacon_c.charge = -2
model.metabolites.itacon_c.compartment = "c"


# In[ ]:


# For the methylation case:
# Add the methylation SAM->SAH reaction to the model, to use as a target
MT_ADDED = cobra.Reaction("MT_ADDED", name="MT_ADDED")  # Methylating the product, target
EX_METH_ADDED = cobra.Reaction("EX_METH_ADDED", name="EX_METH_ADDED") # Sink for the product
model.add_reactions([MT_ADDED, EX_METH_ADDED])
model.reactions.MT_ADDED.reaction = "amet_c --> ahcys_c + meth_c"
model.reactions.MT_ADDED.type = "native"
model.reactions.MT_ADDED.gene_reaction_rule = "targetgene"
model.reactions.EX_METH_ADDED.reaction = "meth_c -->"
model.reactions.EX_METH_ADDED.type = "native"
model.metabolites.meth_c.name = "Methyl group"
model.reactions.ATPM.delete()
model


# # Select reactions to exclude from knock-out
# Make a set of reactions that you do not want to consider for knock-out, as they are spontaneous or diffusion reactions, or reactions that have no gene information. Make sure that the target reaction is not in this set!

# In[ ]:


# Define reactions that should be excluded from knock-out as these will be impossible in practice.

def determine_reactions_to_exclude(model, spontaneous_gene_id=None, ids_not_to_exclude=None):
    """
    Creates a set of reaction ids from the cobra model that should not be suggested as modification by the algorithm.
    This set includes reactions that occur spontaneously, reactions that have no gene information in the model 
    and transport reactions that occur as result of diffusion.
    
    Arguments:
    model: a cobra model
    spontaneous_gene_id: the gene id (string) that spontaneous reactions are assigned to
    ids_not_to_exclude: set of gene ids that should not be in the resulting set
    """
    
    exclude_reaction_ids = set()

    # Exclude reactions for which no gene is available, contain "diffusion" in their description
    # or are assigned to the spontaneous gene id.
    for r in model.reactions:
        if not r.genes or "diffusion" in r.name or spontaneous_gene_id in [g.id for g in r.genes]: 
            exclude_reaction_ids.add(r.id)
            
    # Make sure that the 'ids_not_to_exclude' reactions are not in the set      
    if ids_not_to_exclude is not None:
        for r_id in ids_not_to_exclude:
            exclude_reaction_ids.discard(r_id)

    return exclude_reaction_ids


# In[ ]:


exclude_reaction_ids = determine_reactions_to_exclude(model, "s0001", set(["ITA_ADDED"]))
exclude_reaction_ids


# # Load universal model for knock-ins
# The universal model used is taken from cameo.models.metanetx_universal_model_bigg and converted into BiGG identifiers using the conversion dictionary cameo.data.metanetx.mnx2bigg, as done in the "03 TheAlgorithm EcoliCore2 - methylation" notebook.

# In[105]:


# Load a universal BiGG model from cameo, consisting of 2819 metabolites and 6416 reactions.
univ_model = cobra.io.load_json_model("models/Universal_BiGG_model.json")
univ_model


# In[ ]:


# Print all metabolites and reactions in a model. For debugging/searching:
for m in univ_model.metabolites:
    print(m.id, m.name, m.compartment, sep="\t")
for r in univ_model.reactions:
    print(r.id, r.name, r.reaction, r.compartments, sep="\t")


# # Add relevant universal reactions to the native model

# In[102]:


def fuse_native_with_universal_model(native_model, universal_model, number_of_knockins, include_demands=False):
    """
    Merges the native cobra model with the relevant reactions from the universal_model. 
    The universal reactions that could be fully connected with the specified number of knockins are selected.
    The native reactions are flagged with a type attribute 'native'. Reactions from the universal model are assigned the 'heterologous' type.
    
    Arguments:
    native_model: the native cobra model
    universal_model: the universal cobra model to select knockins from
    include_demands: whether demand reactions from the universal model with matching reactants are added or not
    number_of_knockins: the number of knockins on which to base the selection of relevant universal reactions
    """
    # If no fusion of models is needed, but just labeling all native reactions with .type = "native":
    if number_of_knockins == 0:
        for r in native_model.reactions:
            r.type = "native"
        return native_model
    
    # Not to modify the models given as arguments
    fused_model = native_model.copy()
    reduced_universal_model = universal_model.copy()
    
    # Keep track of native and heterologous reactions by adding a .type attribute to each reaction:
    for r in fused_model.reactions:
        r.type = "native"
        
    # Make a set structure to identify reactions in the universal model that are duplicates of the native reactions
    # reactions = set(     # This set contains representations of all reactions
    #     frozenset(       # This frozenset contains 1 frozenset with all reactants and 1 with all products
    #         frozenset(   # These 2 frozensets contains all the metabolites in either products or reactants
    #             tuple(metabolite_id, coefficient)  # Coefficients are made positive if the reaction is reversible
    #         )                                      # Otherwise, reactant coefficients are negative
    #     )
    # )

    native_reactions_set = set()
    for r in fused_model.reactions:
        product_list = [] # Storing the metabolite ids for all products
        reactant_list = []
        prod_coef_list = [] # Storing the matching coefficients
        reac_coef_list = []
        # Make lists of the product and reactant ids and their coefficients
        for m in r.products:
            product_list.append(m.id)
        for m in r.reactants:
            reactant_list.append(m.id)        
        for c in r.get_coefficients(r.products):
            prod_coef_list.append(c)
        if r.reversibility:
            for c in r.get_coefficients(r.reactants):
                reac_coef_list.append(-c) # Remove negative sign from coefficients
        else:
            for c in r.get_coefficients(r.reactants):
                reac_coef_list.append(c) # Keep the negative sign to indicate direction
        # Zip these lists to make tuples of (metabolite, coefficient) pairs
        productset = frozenset(zip(product_list, prod_coef_list))
        reactantset = frozenset(zip(reactant_list, reac_coef_list))
        reactionset = set()
        reactionset.add(reactantset)
        reactionset.add(productset)
        native_reactions_set.add(frozenset(reactionset))
        #print(native_reactions_set)
    
    # Remove reactions from the universal model that are boundary reactions (if include_demands is not set to True), 
    # or duplicates of the native model used
    reactions_to_remove = set()
    for r in reduced_universal_model.reactions:
        # Create the same structure as above for this reaction. Compare if it is among the native reactions.
        product_list = []
        reactant_list = []
        prod_coef_list = []
        reac_coef_list = []
        # Make lists of the product and reactant ids and their coefficients
        for m in r.products:
            product_list.append(m.id)
        for m in r.reactants:
            reactant_list.append(m.id)        
        for c in r.get_coefficients(r.products):
            prod_coef_list.append(c)
        if r.reversibility:
            for c in r.get_coefficients(r.reactants):
                reac_coef_list.append(-c) # Remove negative sign from coefficients
        else:
            for c in r.get_coefficients(r.reactants):
                reac_coef_list.append(c) # Keep the negative sign to indicate direction
        # Zip these lists to make tuples of (metabolite, coefficient) pairs
        productset = frozenset(zip(product_list, prod_coef_list))
        reactantset = frozenset(zip(reactant_list, reac_coef_list))
        reactionset = set()
        reactionset.add(reactantset)
        reactionset.add(productset)
        
        if (r.products == [] and not include_demands) or                 reactionset in native_reactions_set:
            reactions_to_remove.add(r)

    reduced_universal_model.remove_reactions(reactions_to_remove)
    
    
    # Add the relevant reactions from universal model to the organism's native model, 
    # based on the connection of its metabolites to the native model:
    # For every 2 knock-ins allowed, add all reactions that have either all their reactants or all products present in the model so far.
    # This means that knock-in pathways are build from both start and end, which will connect at some point if enough knock-ins are allowed.
    # Filling the gap from two sides is needed to take reversible reactions into account, which cause problems if building only from the reactant side. 
    # The last knock-in (number_of_knockins = 1) allowed should be connected to the model by both all reactants and products in order to carry flux
    while number_of_knockins > 0:
        model_metabolite_ids = [m.id for m in fused_model.metabolites]
        for r in reduced_universal_model.reactions:
            if (number_of_knockins > 1 and all(m.id in model_metabolite_ids for m in r.reactants) or             (r.products != [] and all(m.id in model_metabolite_ids for m in r.products))) or             (number_of_knockins == 1 and all(m.id in model_metabolite_ids for m in r.metabolites)):    
                    add = r.copy()
                    add.type = "heterologous"
                    fused_model.add_reaction(add)
        number_of_knockins -= 2

               
    # Remove all reactions that can't carry flux in the end, 
    # to trim away all added heterologous reactions that are never relevant as a knock-in.
    blocked_reactions = cameo.flux_analysis.analysis.find_blocked_reactions(fused_model)
    fused_model.remove_reactions([r for r in blocked_reactions])
         
    return fused_model


# In[103]:


fused_model = fuse_native_with_universal_model(model, univ_model, include_demands=False, number_of_knockins=2)
fused_model


# # Specify metabolites for medium additions

# In[ ]:


# Make a list of metabolite ids that can be added to the medium

medium_metabolites = ["ala__L_c", "arg__L_c", "asn__L_c", "asp__L_c", "cys__L_c", "glu__L_c", "gln__L_c", "gly_c", "his__L_c", 
"ile__L_c", "leu__L_c", "lys__L_c", "met__L_c", "phe__L_c", "pro__L_c", "ser__L_c", "thr__L_c", "trp__L_c", 
"tyr__L_c", "val__L_c", "fru_c","mal__L_c", "lac__D_c", "ac_c"]
# 24 (L-)Amino acids, D-fructose, L-malate, D-lactate,  acetate

# For the smallest Ecolicore:
#medium_metabolites = ["glu__L_c", "gln__L_c","lac__D_c", "ac_c"]


# # Create the MILP problem
# This is the main algorithm, creating a MILP problem that when solved returns one or more growth-coupled designs for the target reaction.

# In[ ]:


for reac in model.reactions:
    if reac.lower_bound > 0 or reac.upper_bound < 0:
        print(reac.id, "has a forced flux. This should be removed")


# In[ ]:


import optlang

def setup_growthcoupling_milp(model, target, medium_metabolites, K=2, L=1, M=1, exclude_reaction_ids=set(), remove_blocked=False):
    """
    This function constructs the MILP problem to predict growth-coupled designs with. 
    K is the number of knockouts that are allowed. L is the number of knockins allowed. M is the number of medium additions.
    exclude_reaction_ids takes a list of reaction ids that shouldn't be considered for knockout/in/medium addition 
    (i.e. spontaneous reactions and exchange reactions). These reactions are thus always allowed to have flux within their model bounds, so this should only be used for excluding knockouts.
    """
    # Check model for forced fluxes, which should not be present. The zero solution must be allowed.
    for r in model.reactions:
        if r.lower_bound > 0 or r.upper_bound < 0:
            raise ValueError("%s has a forced flux. Remove the reaction or change its bounds for the function to work." % r.id)
    
    # If a cobra.Reaction is given as target, make the target in just its ID string.
    if isinstance(target, cobra.core.Reaction):
        target = target.id
    
    original_model = model
    model = model.copy()
    
    # Add boundary reactions for these metabolites named "ADD_<metabolitename>" to the model, with bounds (-10,0).
    medium_addition_reactions = []   # List of the addition reactions added to the model
    for m in medium_metabolites:
        met = model.metabolites.get_by_id(m)
        r = model.add_boundary(met, type="medium addition", reaction_id=("ADD_"+m), lb=-10, ub=0)
        r.type = "medium"
        medium_addition_reactions.append(r)
    
    # Create a set of reactions that should not be considered for knockout (from exclude_from_knockouts list of ids)
    excluded_reactions = set()
    for r_id in exclude_reaction_ids:
        excluded_reactions.add(model.reactions.get_by_id(r_id))
    
    
    # Set the solver to Gurobi for the fastest result. Set to CPLEX if Gurobi is not available.
    if "gurobi" in cobra.util.solver.solvers.keys():
        logger.info("Changing solver to Gurobi and tweaking some parameters.")
        if "gurobi_interface" not in model.solver.interface.__name__:
            model.solver = "gurobi"
        # The tolerances are set to the minimum value. This gives maximum precision.
        problem = model.solver.problem
        problem.params.NodeMethod = 1 # primal simplex node relaxation
        problem.params.FeasibilityTol = 1e-9 #If a flux limited to 0 by a constraint, which range around it is still considered the same as 0 > set smallest possible
        problem.params.OptimalityTol = 1e-3 #how sure the solver has to be about this optimum being really the best it has.
        problem.params.IntFeasTol = 1e-9 #If a value is set to an integer, how much may it still vary? > set smallest possible
        problem.params.MIPgapAbs = 1e-9
        problem.params.MIPgap = 1e-9
        #problem.params.TimeLimit = 200 # Use max 200 seconds when called, return best solution after that
        #problem.params.PoolSearchMode = 1 #0 for only finding the optimum, 1 for finding more solutions (but no quality guaranteed), 2 for finding the n best possible solutions 
        #problem.params.PoolSolutions = 10 # Number of solutions kept when finding the optimal solution
        #problem.params.PoolGap = 0.9 # only store solutions within 90% of the optimal objective value

    elif "cplex" in cobra.util.solver.solvers.keys():
        logger.warning("Changing solver to CPLEX, as Gurobi is not available. This may cause a big slowdown and limit options afterwards.")
        if "cplex_interface" not in model.solver.interface.__name__:
            model.solver = "cplex"
        # The tolerances are set to the minimum value. This gives maximum precision.
        problem = model.solver.problem
        problem.parameters.mip.strategy.startalgorithm.set(1) # primal simplex node relaxation
        problem.parameters.simplex.tolerances.feasibility.set(1e-9) #If a flux limited to 0 by a constraint, which range around it is still considered the same as 0 > set smallest possible
        problem.parameters.simplex.tolerances.optimality.set(1e-3) #possibly fine with 1e-3, try if allowed. Is how sure the solver has to be about this optimum being really the best it has.
        problem.parameters.mip.tolerances.integrality.set(1e-9) #If a value is set to an integer, how much may it still vary? > set smallest possible
        problem.parameters.mip.tolerances.absmipgap.set(1e-9)
        problem.parameters.mip.tolerances.mipgap.set(1e-9)
        #problem.parameters.mip.pool.relgap.set(0.9) # For populate: find all solutions within 10% of the optimum for relgap = 0.1
        #problem.parameters.timelimit.set(200) # Use max 200 seconds for solving
        #problem.parameters.mip.limits.populate.set(20) # Find max 20 solutions (=default)
   
    else:
        logger.warning("You are trying to run 'OptCouple' with %s. This might not end well." %
                      model.solver.interface.__name__.split(".")[-1])
    
    # Remove reactions that are blocked: no flux through these reactions possible. This will reduce the search space for the solver, if not done already.
    if remove_blocked:
        blocked_reactions = cameo.flux_analysis.analysis.find_blocked_reactions(fused_model)
        fused_model.remove_reactions(blocked_reactions)
        logger.debug("Removed " + str(len(blocked_reactions)) + " reactions that were blocked")
    
    # Make dual
    model_without_target = model.copy() # This variable looks unnecessary, but is kept out of fear of messing stuff up
    model_without_target.optimize()
    dual_problem = convert_linear_problem_to_dual(model_without_target.solver)
    logger.debug("Dual problem successfully created")

    # Combine primal and dual
    primal_problem = model.solver

    for var in dual_problem.variables:  # All variables in the dual are copied to the primal
        var = primal_problem.interface.Variable.clone(var)
        primal_problem.add(var)
    for const in dual_problem.constraints:  # All constraints in the dual are copied to the primal
        const = primal_problem.interface.Constraint.clone(const, model=primal_problem)
        primal_problem.add(const)
    logger.debug("Dual and primal combined")

    dual_problem.optimize()
    
    
    # Dictionaries to hold the binary control variables: 
    native_y_vars = {}       # 1 for knockout, 0 for active
    heterologous_y_vars = {} # 1 for knockin, 0 for inactive
    medium_y_vars = {}       # 1 for medium addition (up to -10), 0 for no addition
    
    # Now the fun stuff
    constrained_dual_vars = set()
    
    # Fill the dictionaries with binary variables
    # For the knockouts: only for the native reactions which are not exchanges
    for reaction in [r for r in model.reactions - excluded_reactions if r.type == "native"]:

        # Add constraint variables
        interface = model.solver.interface
        y_var = interface.Variable("y_" + reaction.id, type="binary")

        # Constrain the primal: flux through reactions maximum within (-1000, 1000), or smaller boundaries defined before
        model.solver.add(interface.Constraint(reaction.flux_expression - 1000 * (1 - y_var), ub=0, name="primal_y_const_"+reaction.id+"_ub"))
        model.solver.add(interface.Constraint(reaction.flux_expression + 1000 * (1 - y_var), lb=0, name="primal_y_const_"+reaction.id+"_lb")) 

        # Constrain the dual
        constrained_vars = []

        if reaction.upper_bound != 0:
            dual_forward_ub = model.solver.variables["dual_" + reaction.forward_variable.name + "_ub"]
            model.solver.add(interface.Constraint(dual_forward_ub - 1000 * y_var, ub=0))
            constrained_vars.append(dual_forward_ub)
        if reaction.lower_bound != 0:
            dual_reverse_ub = model.solver.variables["dual_" + reaction.reverse_variable.name + "_ub"] 
            model.solver.add(interface.Constraint(dual_reverse_ub - 1000 * y_var, ub=0))
            constrained_vars.append(dual_reverse_ub)
        constrained_dual_vars.update(constrained_vars)

        # Add to dictionary with binary variables for native reactions:
        native_y_vars[y_var] = reaction


    # For the knockins and medium additions:
    for reaction in [r for r in model.reactions - excluded_reactions if r.type == "heterologous" or r.type == "medium"]:          
        # Add constraint variables
        interface = model.solver.interface
        y_var = interface.Variable("y_" + reaction.id, type="binary")

        # Constrain the primal: flux through reactions maximum within (-1000, 1000), or smaller boundaries defined before
        model.solver.add(interface.Constraint(reaction.flux_expression - 1000 * y_var, ub=0, name="primal_y_const_"+reaction.id+"_ub"))
        model.solver.add(interface.Constraint(reaction.flux_expression + 1000 * y_var, lb=0, name="primal_y_const_"+reaction.id+"_lb"))  

        # Constrain the dual
        constrained_vars = []

        if reaction.upper_bound != 0:
            dual_forward_ub = model.solver.variables["dual_" + reaction.forward_variable.name + "_ub"]
            model.solver.add(interface.Constraint(dual_forward_ub - 1000 * (1 - y_var), ub=0))
            constrained_vars.append(dual_forward_ub)
        if reaction.lower_bound != 0:
            dual_reverse_ub = model.solver.variables["dual_" + reaction.reverse_variable.name + "_ub"]
            model.solver.add(interface.Constraint(dual_reverse_ub - 1000 * (1 - y_var), ub=0))
            constrained_vars.append(dual_reverse_ub)
        constrained_dual_vars.update(constrained_vars)

        # Add y variable to the corresponding modifications dictionary
        if reaction.type == "medium":
            medium_y_vars[y_var] = reaction         
        elif reaction.type == "heterologous":
            heterologous_y_vars[y_var] = reaction
            
    logger.debug("Control variables created")
    

    # Set the objective
    primal_objective = model.solver.objective
    dual_objective = interface.Objective.clone(
        dual_problem.objective, model=model.solver
    )

    reduced_expression = optlang.symbolics.Add(
        *((c * v) for v, c in dual_objective.expression.as_coefficients_dict().items()
        if v not in constrained_dual_vars)
    )
    dual_objective = interface.Objective(reduced_expression, direction=dual_objective.direction)

    full_objective = interface.Objective(primal_objective.expression - dual_objective.expression, direction="max")
    model.objective = full_objective
    logger.debug("Objective created")
    
        
    # Add number of knockouts constraint. ub=K+1 as the target will be forced to be 1 as well
    knockout_number_constraint = model.solver.interface.Constraint(
        optlang.symbolics.Add(*native_y_vars), lb=0, ub=K+1, name="number_of_knockouts_constraint"
    )
    model.solver.add(knockout_number_constraint)
   
    # Add number of knockins constraint
    knockin_number_constraint = model.solver.interface.Constraint(
        optlang.symbolics.Add(*heterologous_y_vars), lb=0, ub=L, name="number_of_knockins_constraint"
    )
    model.solver.add(knockin_number_constraint)
  
    # Add number of medium additions constraint
    medium_additions_number_constraint = model.solver.interface.Constraint(
        optlang.symbolics.Add(*medium_y_vars), lb=0, ub=M, name="number_of_medium_additions_constraint"
    )
    model.solver.add(medium_additions_number_constraint)
    
    logger.debug("Added constraint for number of knockouts, knockins and medium additions")

    
    # Force target control variable to be 1, but allow flux in primal
    model.solver.constraints["primal_y_const_" + target + "_lb"].lb = -2000
    model.solver.constraints["primal_y_const_" + target + "_ub"].ub = 2000
    model.variables["y_" + target].lb = 1
    
    return model, native_y_vars, heterologous_y_vars, medium_y_vars


# In[ ]:


milp, native_y_vars, heterologous_y_vars, medium_y_vars = setup_milp(fused_model, "ITA_ADDED",                    medium_metabolites, K=8, L=0, M=1, remove_blocked=False, exclude_reaction_ids=exclude_reaction_ids)


# In[ ]:


milp2, native_y_vars2, heterologous_y_vars2, medium_y_vars2 = setup_milp(fused_model, "MT_ADDED",                    medium_metabolites, K=1, L=2, M=1, remove_blocked=False, exclude_reaction_ids=exclude_reaction_ids)


# In[ ]:


# To see how many reactions apply for modification in each category
print(len(native_y_vars), len(heterologous_y_vars), len(medium_y_vars))


# # Solve the MILP problem to find the optimal solution
# This way of solving only returns the best solution the solver can find. See further below for solving using a solution pool, returning multiple solutions besides the optimal one.

# In[ ]:


# Solve the MILP problem to retrieve the optimal solution

# If speed is important, set a timelimit. Even in a very short time, it will often come up with a solution.
#milp.solver.problem.params.TimeLimit = 200

from timeit import default_timer as timer
start = timer()

milp.optimize()

end = timer()
print("Optimization time:", end - start, "s")
print("Optimization time:", (end - start)/60, "m")
print("Optimization time:", (end - start)/3600, "h")


# In[ ]:


print("MILP objective value:", milp.objective.value)


# In[ ]:


for y in native_y_vars:
    if y.primal > 0.99:
        print(y, y.primal)


# In[ ]:


for y in heterologous_y_vars:
    if y.primal > 0.99:
        print(y, y.primal)


# In[ ]:


for y in medium_y_vars:
    if y.primal > 0.99:
        print(y, y.primal)


# # GUROBI Solution pool search
# Not just retrieve the optimal solution, but get back a variety of solutions (= growth-coupled designs) to explore and choose from.

# In[ ]:


milp.solver.problem.params.PoolSearchMode = 2 #0 for only finding the optimum, 1 for finding more solutions (but no quality guaranteed), 2 for finding the n best possible solutions 
milp.solver.problem.params.PoolSolutions = 10 # Number of solutions kept when finding the optimal solution, max 2,000,000,000
milp.solver.problem.params.PoolGap = 0.99 # only store solutions within 90% of the optimal objective value
milp.solver.problem.params.TimeLimit = 100 # Set a time limit used for finding the solution, in seconds.


# In[ ]:


from timeit import default_timer as timer
start = timer()

milp.optimize()

end = timer()
print("Optimization time:", end - start, "s")
print("Optimization time:", (end - start)/60, "m")
print("Optimization time:", (end - start)/3600, "h")
optimization_time_sec = end - start


# In[ ]:


for v in milp.solver.problem.getVars():
    if v.VarName.startswith("y_") and v.x > 0.99:
        print(v.VarName, v.x)


# In[ ]:


for solution_number in range(milp.solver.problem.solcount):
    milp.solver.problem.params.SolutionNumber = solution_number
    print("Solution number:", solution_number)
    print("Objective:", milp.solver.problem.PoolObjVal)
    for v in milp.solver.problem.getVars():
        if v.VarName.startswith("y_") and v.xn > 0.99:
            reaction_name = v.VarName[2:]
            reaction_type = milp.reactions.get_by_id(reaction_name).type
            print(v.VarName, v.xn, reaction_type, sep= "\t")
    print("\n")


# # Functions for processing and analyzing the Gurobi solution pool

# In[ ]:


def create_runname(milp):
    """
    Create a runname for the current run:
    YYYYMMDD_HHMM-<model name>-<target ID>
    This makes that all runs have a unique name, to be used to label all files produced in the run.
    """
    import datetime
    # Retrieve current time
    date_time = datetime.datetime.now().strftime("%Y%m%d_%H%M")
    # Retrieve target from model, to make sure that this is the correct target 
    # (in contrast to an extra input to this function than can be forgotten to be changed)
    target = [v.VarName[2:] for v in milp.solver.problem.getVars() if v.LB == 1 and v.VarName.startswith("y_")][0]
    runname = "__".join([date_time, milp.id, target])
    return runname

runname = create_runname(milp)


# In[ ]:


def write_parameters_to_file(milp, runname, excluded_reaction_ids, optimization_time_sec):
    filename = "%s__details.txt" % runname
    split_runname = runname.split("__")
    date_time = split_runname[0][:4] + "-" + split_runname[0][4:6] + "-" + split_runname[0][6:8] + " " + split_runname[0][9:11] + ":" + split_runname[0][11:13]
    with open(filename, "w") as output_file:
        output_file.write("Run details for %s:\n" % runname)
        output_file.write("Date: %s\n" % date_time)
        output_file.write("Model: %s\n" % split_runname[1])
        output_file.write("Target: %s\n" % split_runname[2])

        # In case no knockins or medium additions are allowed: set to 0.
        ki_var = milp.solver.problem.getVarByName("number_of_knockins_constraint_aux")
        ki_number = int(ki_var.x) if ki_var is not None else 0
        ma_var = milp.solver.problem.getVarByName("number_of_medium_additions_constraint_aux")
        ma_number = int(ma_var.x) if ma_var is not None else 0    
          
        output_file.write("Knockouts allowed: %i\n" % (int(milp.solver.problem.getVarByName("number_of_knockouts_constraint_aux").x)-1))
        output_file.write("Knockins allowed: %i\n" % ki_number)
        output_file.write("Medium additions allowed: %i\n" % ma_number)
        output_file.write("Number of solutions found: %s\n" % milp.solver.problem.solcount)
        
        solver_settings, solver_parameters = str(milp.solver.problem)[1:-1].split(", Parameter changes: ")
        solver_settings = solver_settings.replace(" Unnamed: ", "\n\t")
        output_file.write("\nSolver settings: \n\t%s\n" % solver_settings)
        output_file.write("Optimization time: %0.0f s, %0.1f m, %0.2f h\n" % (optimization_time_sec, optimization_time_sec/60, optimization_time_sec/3600))
        output_file.write("Solver status: %s\n" % milp.solver.status)
        solver_parameters = solver_parameters.split(", ")
        output_file.write("Parameter changes:\n")
        [output_file.write("\t%s\n" % setting) for setting in solver_parameters]
        output_file.write("\nReactions excluded from modification:\n")
        output_file.write(", ".join(excluded_reaction_ids))
        output_file.write("\n")
                          
write_parameters_to_file(milp, runname, exclude_reaction_ids, optimization_time_sec)


# In[ ]:


def extract_solutions_from_gurobi_pool(milp, target_id, biomass_id):
    """
    Extracts the individual solutions from a MILP solution pool created by the Gurobi solver.
    Returns the solution pool and a header with information to accompany the list of dicts.
    """
    if "gurobi_interface" not in milp.solver.interface.__name__:
        raise ValueError("The solver for the MILP is not Gurobi. This function only works for Gurobi created solution pools.")
    if target_id not in milp.solver.problem.getVars():
        raise ValueError("Specified target reaction ID %s not found in the variable list." % target_id)
    if biomass_id not in milp.solver.problem.getVars():
        raise ValueError("Specified biomass reaction ID %s not found in the variable list." % biomass_id)
        
    solution_pool = list()
    for solution_number in range(milp.solver.problem.solcount):
        solution = dict()
        milp.solver.problem.params.SolutionNumber = solution_number
        solution["number"] = solution_number
        # Note that the objective is the growth coupling potential, not the growth or production rate
        solution["objective"] = milp.solver.problem.PoolObjVal 
        
        # Prepare dict for looping over variables
        # Assigning all keys here to put them in a certain order in the dict
        target_flux_name = "target_flux_(%s)" % target_id
        biomass_id_reverse = biomass_id + "_reverse_"
        target_id_reverse = target_id + "_reverse_"
        
        solution["growth_coupling_slope"] = None
        solution[target_flux_name] = 0
        solution["growth_rate"] = 0
        solution["growth_without_target"] = None
        #solution["target/biomass"] = None
        solution["knockout"] = list()
        solution["knockin"] = list()
        solution["medium_addition"] = list()

        
        for v in milp.solver.problem.getVars():
            if v.VarName.startswith("y_") and v.xn > 0.99:
                reaction_id = v.VarName[2:] #removing "y_"
                reaction_type = milp.reactions.get_by_id(reaction_id).type
                if reaction_type == "native" and reaction_id != target_id:
                    solution["knockout"].append(reaction_id)
                elif reaction_type == "heterologous":
                    solution["knockin"].append(reaction_id)
                elif reaction_type == "medium":
                    solution["medium_addition"].append(reaction_id)
            
            # Extract target flux and growth rate. Subtract the forward and reverse fluxes to get the nett flux.
            if v.VarName == target_id:
                solution[target_flux_name] += v.xn
                # In case target_id or biomass_id are not found, and these values are assigned None to start with:
                #solution["target_flux"] = v.xn if solution["target_flux"] is None else solution["target_flux"] + v.xn
            if v.VarName.startswith(target_id_reverse):
                solution[target_flux_name] -= v.xn
            if v.VarName == biomass_id:
                solution["growth_rate"] += v.xn
            if v.VarName.startswith(biomass_id_reverse):
                solution["growth_rate"] -= v.xn
          
        solution["growth_without_target"] = solution["growth_rate"] - solution["objective"]
        if round(solution["objective"],7) > 0:
            solution["growth_coupling_slope"] = solution[target_flux_name] / solution["objective"]
            
        solution_pool.append(solution)
        
        
    return solution_pool

solution_pool = extract_solutions_from_gurobi_pool(milp, "MT_ADDED", "Ec_biomass_iJO1366_core_53p95M")


# In[ ]:


def write_solution_pool_to_csv(solution_pool, runname, suffix=""):
    """Write the solution pool list of dictionaries to csv"""
    # Alternatively, numpy could be used to do this.
    import unicodecsv as csv
    filename = "%s__solution_pool%s.csv" % (runname, suffix)
    keys = solution_pool[0].keys()
    with open(filename, 'wb') as output_file:
        dict_writer = csv.DictWriter(output_file, keys)
        dict_writer.writeheader()
        dict_writer.writerows(solution_pool)
        
write_solution_pool_to_csv(minimal_solution_pool, runname, suffix="new")


# In[ ]:


def write_solution_pool_to_json(solution_pool, runname, suffix="", indent=4):
    """Write the list of dicts to json format for easy retrieval of the data structure"""
    import json
    filename = "%s__solution_pool%s.json" % (runname, suffix)
    with open(filename, "w") as output_file:
        json.dump(solution_pool, output_file, indent=indent)
        
write_solution_pool_to_json(minimal_solution_pool, runname, suffix="new")


# In[ ]:


from itertools import chain, combinations

# Function used to make all possible modification combinations in select_minimal_designs()
def powerset(iterable):
    """
    From the itertools docs:
    Function that creates all possible combination sets out of a list of elements.
    i.e. powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))


# In[ ]:


def select_minimal_designs(native_model, milp, solution_pool, target_id=None, biomass_id=None):
    """
    Selects the minimal set of modifications from the solution pool, 
    e.g. filters out solutions that have additional modifications that do not contribute to the growth coupling
    """
    
    #Autodetect target and biomass IDs if not given:
    if target_id is None:
        target_id = [v.VarName[2:] for v in milp.solver.problem.getVars() if v.LB == 1 and v.VarName.startswith("y_")][0]
    if biomass_id is None:
        biomass_id = [r.id for r in model.reactions if "biomass" in r.id.lower()][0]
    
    target_flux_name = "target_flux_(%s)" % target_id
    minimal_solution_pool = list()
    minimal_designs_so_far = list()
    
    # Loop over all solutions in the solution pool
    for solution in solution_pool:
        ko = solution["knockout"]
        ki = solution["knockin"]
        ma = solution["medium_addition"]      
        new_reactions = ki + ma
        modifications = ko + ki + ma
        
        # Verify whether a subset of the design is already in the minimal solution pool so far
        # Note that this is a bit of a shortcut, as it doesn't check whether the growth-coupling slope/strength matches as well
        # So designs with a stronger growth-coupling than found earlier could be discarded here.
        # Checking this is not strictly needed, but greatly improves speed. Otherwise, just discard the next 7 lines.
        already_found = False
        for design in minimal_designs_so_far:
            if all(reaction in modifications for reaction in design):
                already_found = True
                break
        
        if not already_found:
            design_found = False
            minimal_combination = dict()
            with native_model as model:
                # Add the extra reactions in this solution to the native model
                for r in new_reactions:
                    model.add_reaction(milp.reactions.get_by_id(r).copy())
                # Loop over all combinations that these genes can be allowed to have flux
                # starting with the individual genes, then pairs, etc.
                powerset_combinations = list(powerset(modifications))
                for i in range(len(powerset_combinations)):
                    combination = powerset_combinations[i]
                    with model as test_model:
                        for reaction in ko: 
                            if reaction in combination:
                                test_model.reactions.get_by_id(reaction).knock_out()
                        for reaction in new_reactions: 
                            if reaction not in combination:
                                test_model.reactions.get_by_id(reaction).knock_out()
                        test_model.optimize()
                        target_flux = test_model.reactions.get_by_id(target_id).flux
                        biomass_flux_max = test_model.reactions.get_by_id(biomass_id).flux
                        test_model.reactions.get_by_id(target_id).bounds = (0, 0)
                        biomass_flux_min = test_model.slim_optimize()
                        biomass_flux_delta = biomass_flux_max - biomass_flux_min
                        if (round(biomass_flux_delta,5) != 0 and round(target_flux/biomass_flux_delta,2) == round(solution["growth_coupling_slope"],2)) or i == len(powerset_combinations) - 1:
                            minimal_combination["objective"] = round(biomass_flux_delta,7)
                            minimal_combination["growth_coupling_slope"] = round(target_flux/biomass_flux_delta,7)
                            minimal_combination[target_flux_name] = round(target_flux,7)
                            minimal_combination["growth_rate"] = round(biomass_flux_max,7)
                            minimal_combination["growth_without_target"] = round(biomass_flux_min,7)
                            minimal_combination["knockout"] = [r for r in ko if r in combination]
                            minimal_combination["knockin"] = [r for r in ki if r in combination]
                            minimal_combination["medium_addition"] = [r for r in ma if r in combination]
                            design_found = True
                            break

            if design_found:
                # Only keep minimal solutions that are unique
                if minimal_solution_pool == list():
                    minimal_solution_pool.append(minimal_combination)
                    minimal_designs_so_far.append(minimal_combination["knockout"]+minimal_combination["knockin"]+minimal_combination["medium_addition"])
                else:
                    for key in minimal_combination.keys():
                        if minimal_combination[key] not in [value for i in range(len(minimal_solution_pool))                                                       for value in minimal_solution_pool[i].values()]:
                            minimal_solution_pool.append(minimal_combination)
                            minimal_designs_so_far.append(minimal_combination["knockout"]+minimal_combination["knockin"]+minimal_combination["medium_addition"])
                            break
            else:
                print("Error: No design found for prediction: %s" % ", ".join(modifications))
                # This is the case for designs that have an abberant objective value due to an error in the algorithm
                # You could also decide to add them all to the minimal solution pool as long as the error is not resolved.
                    
    return minimal_solution_pool


minimal_solution_pool = select_minimal_designs(model, milp, solution_pool, "MT_ADDED", "Ec_biomass_iJO1366_core_53p95M")
minimal_solution_pool


# In[ ]:


def write_solution_pool_to_verbose_txt(solution_pool, milp, runname, suffix):
    """
    Make a more easily interpretable file with all solutions in the solution pool, containing for each proposed design
    the objective value, target flux, growth rate with and without the target reaction, 
    as well as the ids and equations of the reactions modified. Moreover, all metabolite names in these are spelled out.
    """
    # Automatically retrieve the target reaction ID
    target_id = [v.VarName[2:] for v in milp.solver.problem.getVars() if v.LB == 1 and v.VarName.startswith("y_")][0]
    
    with open("%s__solution_pool%s.txt" % (runname, suffix), "w") as output_file:
        output_file.write("Solution pool for run %s:\n" % runname)
        for i, solution in enumerate(solution_pool):
            output_file.write("\n\n>>> Solution %i <<<\n" % i)
            output_file.write("Target flux: %f    Growth rate: %f\n" % (solution["target_flux_(%s)" % target_id], solution["growth_rate"]))
            output_file.write("Growth-coupling slope: %f    Growth without target: %f\n" % (solution["growth_coupling_slope"], solution["growth_without_target"]))
            if solution["knockout"] != []:
                output_file.write("\n> Knockout:\n")
                for r_id in solution["knockout"]:
                    ko = milp.reactions.get_by_id(r_id)
                    output_file.write("%s (%s):\n    %s   %s\n" % (r_id, ko.name, ko.reaction, ko.bounds))
                    for m in ko.reactants + ko.products:
                        output_file.write("        %s: %s\n" % (m.id, m.name))
            if solution["knockin"] != []:
                output_file.write("\n> Knockin:\n")
                for r_id in solution["knockin"]:
                    ki = milp.reactions.get_by_id(r_id)
                    EC_string = " EC: %s" % ki.annotation.get("EC").replace(";", "; ") if ki.annotation.get("EC") is not None else ""
                    output_file.write("%s (%s, %s):\n    %s   %s\n" % (r_id, ki.name, EC_string, ki.reaction, ki.bounds))
                    for m in ki.reactants + ki.products:
                        output_file.write("        %s: %s\n" % (m.id, m.name))
            if solution["medium_addition"] != []:
                output_file.write("\n> Add to the medium:\n")
                for r_id in solution["medium_addition"]:
                    ma = milp.metabolites.get_by_id(r_id[4:])
                    output_file.write("%s: %s (%s)\n" % (r_id, ma.name, ma.formula))
                
write_solution_pool_to_verbose_txt(minimal_solution_pool, milp, runname, "minimal")


# In[ ]:


import numpy as np
import matplotlib
#matplotlib.use('Agg') # For on HPC, to use backend for saving files and not needing to display them. 
                       # Load this before everything else in the start, as other libraries otherwise may load matplotlib first.
from matplotlib import pyplot as plt
from matplotlib.patches import Polygon
from matplotlib.ticker import AutoMinorLocator

def create_production_envelope(orig_model, target_id, biomass_id, plottitle="Production envelope", filename="productionenvelope.png", xmax=None, ymax=None, points=100):
    """Creates a production envelope for a metabolic model using matplotlib pyplot's Polygon."""
    with orig_model as model:
        model.objective = biomass_id
        model.objective_direction = "max"
        max_growth_rate = model.slim_optimize()
        growth_rate_interval = [max_growth_rate/(points-1) * i for i in range(0,points)]
        model.objective = target_id
        max_target_flux = []
        min_target_flux = []
        for growth_rate in growth_rate_interval:
            model.reactions.get_by_id(biomass_id).bounds = (growth_rate, growth_rate)
            model.objective_direction = "max"
            max_target_flux.append(model.slim_optimize())
            model.objective_direction = "min"
            min_target_flux.append(model.slim_optimize())        
    # Make one list of tuples out of the minimal and maximal values
    max_target_flux.extend(min_target_flux[::-1])
    growth_rate_interval.extend(growth_rate_interval[::-1])
    ppp = list(zip(growth_rate_interval, max_target_flux))
    # Make into numpy array
    dt=np.dtype('float,float')
    np_ppp = np.array(ppp,dtype=dt)
    # Plot with matplotlip pyplot

    fig = plt.figure(10, figsize=(5,3.5), dpi=300)
    ax = fig.add_subplot(111)
    patch = Polygon(ppp, edgecolor="#fc9402" , facecolor="#ffe3b7", linewidth=2, closed=True)
    ax.add_patch(patch)
    
    if xmax is None:
        xmax = max_growth_rate * 1.14
    if ymax is None:
        ymax = max(max_target_flux) * 1.2
    
    xrange = [0, xmax]
    yrange = [0, ymax]
    ax.set_xlim(*xrange)
    ax.set_ylim(*yrange)
    plt.xlabel(r'%s (h$^{-1}$)'% biomass_id, fontsize=8)
    plt.ylabel(r'%s (mmol gDW$^{-1}$ h$^{-1}$)'% target_id, fontsize=8)
    ax.set_title(" ")
    plt.grid(b=True, which='major', axis='both', alpha=0.2)
    minorLocator = AutoMinorLocator()
    ax.xaxis.set_minor_locator(minorLocator)
    minorLocator = AutoMinorLocator()
    ax.yaxis.set_minor_locator(minorLocator)
    ax.text(0, 1.02, "%s" % plottitle, fontsize=9,
        horizontalalignment='left',
        verticalalignment='bottom',
        transform=ax.transAxes)
    plt.tight_layout()
    plt.savefig("%s" % filename)
    plt.close(fig)
    
create_production_envelope(model, "MT_ADDED", "Ec_biomass_iJO1366_core_53p95M", "Production envelope")


# In[ ]:


import os
def create_solution_pool_production_envelopes(native_model, milp, solution_pool, target_id, biomass_id, runname):
    """
    Creates production envelopes for all solutions in the solution pool.
    Stores them as png in a separate folder matching the runname.
    """
    # Create a folder to put all envelopes in
    dirname = "%s_prod_envs" % runname
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    
    # Make a production envelope of the native model to start with
    filename = "%s/%s__%s__no_modifications" % (dirname, native_model.id, target_id)
    xmax = native_model.slim_optimize() * 1.14
    create_production_envelope(native_model, target_id, biomass_id, plottitle="%s with %s target" % (native_model.id, target_id), filename=filename, xmax=xmax)

    # Loop over all solution, modify the model and make a production envelope
    for i, solution in enumerate(solution_pool):
        with native_model as model:
            for ko in solution["knockout"]:
                model.reactions.get_by_id(ko).knock_out()
            for ki in solution["knockin"] + solution["medium_addition"]:
                new = milp.reactions.get_by_id(ki).copy()
                model.add_reaction(new)
            filename = "%s/%s__%s__solution_%i" % (dirname, model.id, target_id, i)
            plottitle = "%s with " % model.id
            if solution["knockout"] != []:
                plottitle += "KO: %s; " % ", ".join(solution["knockout"])
            if solution["knockin"] != []:
                plottitle += "KI: %s; " % ", ".join(solution["knockin"])    
            if solution["medium_addition"] != []:
                plottitle += "ADD: %s; " % ", ".join([r_id[4:] for r_id in solution["medium_addition"]])
            
            create_production_envelope(model, target_id, biomass_id, plottitle=plottitle, filename=filename, xmax=xmax)

        
#create_solution_pool_production_envelopes(model, milp, minimal_solution_pool, "MT_ADDED", "BIOMASS_Ec_iJO1366_core_53p95M", runname)
create_solution_pool_production_envelopes(model, milp, solution_pool, "ITA_ADDED", "BIOMASS_Ec_iJO1366_core_53p95M", runname)


# In[ ]:


# Example for the methylation case on how to make a production envelope for a single design,
# by first implementing the modifications and then calling the create_production_envelope() function
with model as methmodel:
    # CFB labdesign for methylation
    methmodel.reactions.SERAT.knock_out()
    methmodel.add_reactions([univ_model.reactions.CYSTGL.copy(), univ_model.reactions.CYSTS.copy()])
    met = methmodel.metabolites.met__L_c
    methmodel.add_boundary(metabolite=met, type="medium addition", lb=-10, ub=0, reaction_id="ADD_met__L_c")
    methmodel.optimize()
    target_flux = methmodel.reactions.get_by_id("MT_ADDED").flux
    biomass_flux_max = methmodel.reactions.get_by_id("Ec_biomass_iJO1366_core_53p95M").flux
    methmodel.reactions.get_by_id("MT_ADDED").bounds = (0, 0)
    biomass_flux_min = methmodel.slim_optimize()
    biomass_flux_delta = biomass_flux_max - biomass_flux_min
    print("Target flux: %f;\nGrowth rate: %f;\nWithout MT_ADDED: %f;\nObjective: %f" %(target_flux, biomass_flux_max, biomass_flux_min, biomass_flux_delta))
    create_production_envelope(methmodel, "MT_ADDED","Ec_biomass_iJO1366_core_53p95M", filename="ECC2__MT_ADDED__CFBlabdesign_test", plottitle="ECC2 with KO: SERAT; KI: CYSTGL, CYSTS; ADD: met__L_c" )


# # CPLEX Populate method to create multiple solutions
# 
# Although it turned out that Gurobi is a much faster solver, first the solution pool option and the retrieval of its solutions in cplex was explored. This part of the code was kept in case the use of this solver would be explored in the future, but can otherwise be ignored.
# 
# Note that this can use the optimized MILP as a starting point, so it doesn't need to go through that first stage anymore. If no optimized model is provided, it will optimize first and then look for other solutions within the allowed objective range.

# In[ ]:


milp.solver.problem.parameters.mip.pool.absgap.set(0.90)
milp.solver.problem.parameters.mip.pool.relgap.set(0.99)
milp.solver.problem.parameters.timelimit.set(240) # Use max 240 seconds
milp.solver.problem.parameters.mip.limits.populate.set(50) # Find max 50 solutions
milp.solver.problem.parameters.mip.pool.intensity.set(2)


# In[ ]:


milp.solver.problem.populate_solution_pool()


# In[ ]:


# Most optimal solution found (incumbent)
print(milp.solver.problem.solution.get_objective_value())

# Number of solutions in the pool
number_of_solutions_in_pool = milp.solver.problem.solution.pool.get_num()
print(number_of_solutions_in_pool)

# Number of solutions which were discarded due to a too big difference between the solution and the optimal solution (the relgap parameter)
print(milp.solver.problem.solution.pool.get_num_replaced())

# Mean objective value of all solutions in the pool
print("Mean objective value:", milp.solver.problem.solution.pool.get_mean_objective_value())

# Names of each solution in the pool
pool_names = milp.solver.problem.solution.pool.get_names()


# In[ ]:


# Get a list of the names of all variables/constraints in the problem:
variable_names = milp.solver.problem.variables.get_names()
number_of_milp_variables = len(variable_names)

# Create a boolean list indicating whether a constraint index corresponds to a y_var or not
y_var_constraint_index = []
for i in range(number_of_milp_variables):
    y_var_constraint_index.append(variable_names[i].startswith("y_"))


# In[ ]:


# For each solution, print the name, the objective value and the y_vars that are 1 (meaning: change)
# TODO: how to sort these in native/heterologous/medium? > medium is obvious with 'ADD_', but the others...?
for i in range(number_of_solutions_in_pool):        
    objective_value_i = milp.solver.problem.solution.pool.get_objective_value(i)
    values_i = milp.solver.problem.solution.pool.get_values(i)
    print("Solution:",pool_names[i], "\nObjective-value:", objective_value_i)
    for j in range(number_of_milp_variables):
        if y_var_constraint_index[j] and round(values_i[j], 7) == 1:
            print( variable_names[j], values_i[j])
    
    
    


# In[ ]:


print(milp.solver.problem)


# **Solutions found by 1 run with the following parameters:**
# 
# 
# `
# milp.solver.problem.parameters.mip.pool.absgap.set(0.90)
# milp.solver.problem.parameters.mip.pool.relgap.set(0.99)
# milp.solver.problem.parameters.tune.timelimit.set(600)
# milp.solver.problem.parameters.mip.limits.populate.set(50)
# milp.solver.problem.parameters.mip.pool.intensity.set(2)
# `
# 
# 
# (After setting these parameters, first calling optimize, then populate)
# 
# **Suggested designs:**
# 
# 
# ```
# Solution: p10 
# Objective-value: 0.289890083345977
# y_ALCD2x 1.0
# y_GLUDy 1.0
# y_GLUSy 1.0
# y_NADH16 1.0
# y_ITA_ADDED 1.0
# y_ADD_gln__L_c 1.0
# 
# Solution: p2 
# Objective-value: 0.0030392889852107197
# y_ALCD2x 1.0
# y_GLUDy 1.0
# y_GLUt2r 1.0
# y_NADH16 1.0
# y_ITA_ADDED 1.0
# y_ADD_glu__L_c 1.0
# 
# Solution: p3 
# Objective-value: 0.28893819387560216
# y_ALCD2x 1.0
# y_GLUDy 1.0
# y_GLUSy 1.0
# y_NADH16 1.0
# y_ITA_ADDED 1.0
# y_ADD_glu__L_c 1.0
# 
# Solution: p5 
# Objective-value: 0.2898900833459735
# y_ACALD 1.0
# y_GLUDy 1.0
# y_GLUSy 1.0
# y_NADH16 1.0
# y_ITA_ADDED 1.0
# y_ADD_gln__L_c 1.0
# 
# Solution: p6 
# Objective-value: 0.01645466076876917
# y_ATPS4r 1.0
# y_CYTBD 1.0
# y_GLUDy 1.0
# y_PFL 1.0
# y_ITA_ADDED 1.0
# y_ADD_glu__L_c 1.0
# 
# Solution: p7 
# Objective-value: 0.016454660768769502
# y_ATPS4r 1.0
# y_CYTBD 1.0
# y_FORt 1.0
# y_GLUDy 1.0
# y_ITA_ADDED 1.0
# y_ADD_glu__L_c 1.0
# 
# Solution: p8 
# Objective-value: 0.006784680094342854
# y_CYTBD 1.0
# y_GLUDy 1.0
# y_GLUt2r 1.0
# y_PFL 1.0
# y_ITA_ADDED 1.0
# y_ADD_glu__L_c 1.0
# 
# Solution: p9 
# Objective-value: 0.006784680094342854
# y_CYTBD 1.0
# y_FORt 1.0
# y_GLUDy 1.0
# y_GLUt2r 1.0
# y_ITA_ADDED 1.0
# y_ADD_glu__L_c 1.0
# 
# Solution: p1 
# Objective-value: 0.0030392889852082217
# y_ACALD 1.0
# y_GLUDy 1.0
# y_GLUt2r 1.0
# y_NADH16 1.0
# y_ITA_ADDED 1.0
# y_ADD_glu__L_c 1.0
# ```
