# Script to make a solution pool of growth-coupled strain design predictions
# Created: 01 May 2018
# Author: Valentijn Broeken

# Loading dependencies
import os
os.environ["OPTLANG_USE_SYMENGINE"] = "False"
import matplotlib
matplotlib.use('agg')
import cameo
import cobra
from cobra.test import create_test_model
from optlang.duality import convert_linear_problem_to_dual
import logging
from cameo.models import universal
from cameo.data import metanetx

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()

# Show versions of dependencies
cobra.show_versions()

# Load EColiCore2 model (Hädicke and Klamt, 2017)
# Note that the forBIO_c and glxBIO_c metabolites are fused in this model with for_c and glx_c to fit the BiGG standard and comply with the universal model
model = cobra.io.load_json_model("EColiCore2_from_iJO1366.json")

# Specify reaction names and parameters used throughout the script:
target_id = "ITA_ADDED"
biomass_id = "BIOMASS_Ec_iJO1366_core_53p95M"
number_of_knockouts = 6
number_of_knockins = 0
number_of_medium_additions = 1

# Add target reaction cadA + the metabolite and the exporter for itaconic acid
ITA_ADDED = cobra.Reaction("ITA_ADDED", name="cis-Aconitate decarboxylase")
EX_ITA = cobra.Reaction("EX_ITA", name="Itaconic acid exporter")
model.add_reactions([ITA_ADDED, EX_ITA])
model.reactions.ITA_ADDED.reaction = "acon_C_c + h_c -> itacon_c + co2_c"
model.reactions.EX_ITA.reaction = "itacon_c -->"
model.reactions.ITA_ADDED.gene_reaction_rule = "cadA"
model.metabolites.itacon_c.name = "Itaconic acid"
model.metabolites.itacon_c.formula = "C5H4O4"
model.metabolites.itacon_c.charge = -2
model.metabolites.itacon_c.compartment = "c"

# Set the solver to CPLEX and calculate the baseline growth rate
model.solver = "cplex"
model.objective = biomass_id
model.objective.direction = "max"
print("Growth rate of the original %s model:" % model.id, model.optimize())
# Remove the ATP maintenance reaction to make the 0 solution possible,
# needed for the algorithm to work.
model.remove_reactions([model.reactions.ATPM])
print("Growth rate after removing the ATPM reaction:", model.optimize())

# Define reactions that should be excluded from knock-out as these will be impossible in practice.
def determine_reactions_to_exclude(model, spontaneous_gene_id=None, ids_not_to_exclude=None):
    """
    Creates a set of reaction ids from the cobra model that should not be suggested as modification by the algorithm.
    This set includes reactions that occur spontaneously, reactions that have no gene information in the model 
    and transport reactions that occur as result of diffusion.
    
    Arguments:
    model: a cobra model
    spontaneous_gene_id: the gene id (string) that spontaneous reactions are assigned to
    ids_not_to_exclude: set of gene ids that should not be in the resulting set
    """
    
    exclude_reaction_ids = set()

    # Exclude reactions for which no gene is available, contain "diffusion" in their description
    # or are assigned to the spontaneous gene id.
    for r in model.reactions:
        if not r.genes or "diffusion" in r.name or spontaneous_gene_id in [g.id for g in r.genes]: 
            exclude_reaction_ids.add(r.id)
            
    # Make sure that the 'ids_not_to_exclude' reactions are not in the set      
    if ids_not_to_exclude is not None:
        for r_id in ids_not_to_exclude:
            exclude_reaction_ids.discard(r_id)

    return exclude_reaction_ids

exclude_reaction_ids = determine_reactions_to_exclude(model, "s0001", set([target_id]))


# Load the universal model to pick the knock-in reactions from
universal_model = cobra.io.load_json_model("Universal_BiGG_model.json")


def fuse_native_with_universal_model(native_model, universal_model, number_of_knockins, include_demands=False):
    """
    Merges the native cobra model with the relevant reactions from the universal_model. 
    The universal reactions that could be fully connected with the specified number of knockins are selected.
    The native reactions are flagged with a type attribute 'native'. Reactions from the universal model are assigned the 'heterologous' type.
    
    Arguments:
    native_model: the native cobra model
    universal_model: the universal cobra model to select knockins from
    include_demands: whether demand reactions from the universal model with matching reactants are added or not
    number_of_knockins: the number of knockins on which to base the selection of relevant universal reactions
    """
    # If no fusion of models is needed, but just labeling all native reactions with .type = "native":
    if number_of_knockins == 0:
        for r in native_model.reactions:
            r.type = "native"
        return native_model
    
    
    # Make a list of all reaction strings, used to check if a reaction in the universal model is identical to the native model
    model_reaction_strings = [r.reaction for r in native_model.reactions]
    
    # Make a list of split reaction strings for reversible reactions, in which the reactant and product part of the string
    # are separated and then appended as one list element in the order ["products", "reactants"].
    # Each element in this list thus consists of a list of 2 strings, fx ['h2o_c + pep_c', '2pg_c'] for the reaction 2pg_c <=> h2o_c + pep_c
    model_split_reaction_strings = []
    for r in native_model.reactions:
        if r.reversibility:
            model_split_reaction_strings.append(r.reaction.split(" <=> ")[::-1])
    
    # Not to modify the models given as arguments
    fused_model = native_model.copy()
    reduced_universal_model = universal_model.copy()
    
    # Keep track of native and heterologous reactions by adding a .type attribute to each reaction:
    for r in fused_model.reactions:
        r.type = "native"
        
    # Remove reactions from the universal model that are boundary reactions (if include_demands is not set to True), 
    # or duplicates of the native model used
    i = 0
    while i < len(reduced_universal_model.reactions):
        r = reduced_universal_model.reactions[i]
        if (r.products == [] and not include_demands) or \
        r.reaction in model_reaction_strings or \
        r.reaction.split(" <=> ") in model_split_reaction_strings: # to remove reversible reactions that have products <=> reactants flipped. 
        # ^ Only holds if metabolites are sorted alphabetically, which they normally are.
            r.remove_from_model()
        else:
            i += 1
            
    # Merge the relevant reactions from universal model with the organism's native model:
    # All reactions that have either all their reactants or all products already present in the model so far.
    # Loops once for every two knock-ins allowed, as working from both the product and reactant side
    # Last knock-in allowed should be connected to the model by both all reactants and products in order to carry flux
    while number_of_knockins > 0:
        model_metabolite_ids = [m.id for m in fused_model.metabolites]
        for r in reduced_universal_model.reactions:
            if (number_of_knockins > 1 and all(m.id in model_metabolite_ids for m in r.reactants) or \
            (r.products != [] and all(m.id in model_metabolite_ids for m in r.products))) or \
            (number_of_knockins == 1 and all(m.id in model_metabolite_ids for m in r.metabolites)):    
                    add = r.copy()
                    add.type = "heterologous"
                    fused_model.add_reaction(add)
        number_of_knockins -= 2

               
    # Remove all added heterologous reactions that can't carry flux in the end:
    blocked_reactions = cameo.flux_analysis.analysis.find_blocked_reactions(fused_model)
    [r.remove_from_model() for r in blocked_reactions if r.type == "heterologous"]
         
    return fused_model
    
# Select only universal reactions that could carry flux in the native model with the number of knock-ins allowed.
# Add them to the native model while separating reactions with the .type attribute (native/heterologous)   
fused_model = fuse_native_with_universal_model(model, universal_model, number_of_knockins, include_demands=False)

# Define a list of metabolites that could be added to the medium
# 20 amino acids, fructose, maltose, lactose and acetate
medium_metabolites = ["ala__L_c", "arg__L_c", "asn__L_c", "asp__L_c", "cys__L_c", "glu__L_c", "gln__L_c", "gly_c", "his__L_c", 
"ile__L_c", "leu__L_c", "lys__L_c", "met__L_c", "phe__L_c", "pro__L_c", "ser__L_c", "thr__L_c", "trp__L_c", 
"tyr__L_c", "val__L_c", "fru_c", "mal__L_c", "lac__D_c", "ac_c"]


import optlang

def setup_milp(model, target, medium_metabolites, K=2, L=1, M=1, remove_blocked=False, exclude_reaction_ids=set()):
    """
    This function constructs the MILP. Currently only knockouts are
    implemented. K is the number of knockouts that are allowed.
    L is the number of knockins allowed. M is the number of medium additions.
    exclude_reaction_ids takes a list of reaction ids that shouldn't be considered for knockout/in/medium addition 
    (i.e. spontaneous reactions and exchange reactions). These reactions are thus always allowed to have flux within their model bounds.
    """
    # Check model for forced fluxes, which should not be present. The zero solution must be allowed.
    for r in model.reactions:
        if r.lower_bound > 0 or r.upper_bound < 0:
            raise ValueError("%s has a forced flux. Remove the reaction or change its bounds for the function to work." % r.id)
    
    # If a cobra.Reaction is given as target, make the target in just its ID string.
    if isinstance(target, cobra.core.Reaction):
        target = target.id
    
    original_model = model
    model = model.copy()
    
    # Add boundary reactions for these metabolites named "ADD_<metabolitename>" to the model, with bounds (-10,0).
    medium_addition_reactions = []   # List of the addition reactions added to the model
    for m in medium_metabolites:
        try:
            met = model.metabolites.get_by_id(m)
            r = model.add_boundary(met, type="medium addition", reaction_id=("ADD_"+m), lb=-10, ub=0)
            r.type = "medium"
            medium_addition_reactions.append(r)
        except KeyError as err:
            print("No metabolite found called %s to add to the medium. This metabolite is omitted." % err)
    
    # Create a set of reactions that should not be considered for knockout (from exclude_from_knockouts list of ids)
    excluded_reactions = set()
    for r_id in exclude_reaction_ids:
        excluded_reactions.add(model.reactions.get_by_id(r_id))
    
    
    # Set the solver to Gurobi for the fastest result. Set to CPLEX if Gurobi is not available.
    if "gurobi" in cobra.util.solver.solvers.keys():
        logger.info("Changing solver to Gurobi and tweaking some parameters.")
        if "gurobi_interface" not in model.solver.interface.__name__:
            model.solver = "gurobi"
        # The tolerances are set to the minimum value. This gives maximum precision.
        problem = model.solver.problem
        problem.params.NodeMethod = 1 # primal simplex node relaxation
        problem.params.FeasibilityTol = 1e-9 #If a flux limited to 0 by a constraint, which range around it is still considered the same as 0 > set smallest possible
        problem.params.OptimalityTol = 1e-3 #how sure the solver has to be about this optimum being really the best it has.
        problem.params.IntFeasTol = 1e-9 #If a value is set to an integer, how much may it still vary? > set smallest possible
        problem.params.MIPgapAbs = 1e-9
        problem.params.MIPgap = 1e-9
        #problem.params.TimeLimit = 200 # Use max 200 seconds when called, return best solution after that
        #problem.params.PoolSearchMode = 1 #0 for only finding the optimum, 1 for finding more solutions (but no quality guaranteed), 2 for finding the n best possible solutions 
        #problem.params.PoolSolutions = 10 # Number of solutions kept when finding the optimal solution
        #problem.params.PoolGap = 0.9 # only store solutions within 90% of the optimal objective value

    elif "cplex" in cobra.util.solver.solvers.keys():
        logger.warning("Changing solver to CPLEX, as Gurobi is not available. This may cause a big slowdown and limit options afterwards.")
        if "cplex_interface" not in model.solver.interface.__name__:
            model.solver = "cplex"
        # The tolerances are set to the minimum value. This gives maximum precision.
        problem = model.solver.problem
        problem.parameters.mip.strategy.startalgorithm.set(1) # primal simplex node relaxation
        problem.parameters.simplex.tolerances.feasibility.set(1e-9) #If a flux limited to 0 by a constraint, which range around it is still considered the same as 0 > set smallest possible
        problem.parameters.simplex.tolerances.optimality.set(1e-3) #possibly fine with 1e-3, try if allowed. Is how sure the solver has to be about this optimum being really the best it has.
        problem.parameters.mip.tolerances.integrality.set(1e-9) #If a value is set to an integer, how much may it still vary? > set smallest possible
        problem.parameters.mip.tolerances.absmipgap.set(1e-9)
        problem.parameters.mip.tolerances.mipgap.set(1e-9)
        #problem.parameters.mip.pool.relgap.set(0.9) # For populate: find all solutions within 10% of the optimum for relgap = 0.1
        #problem.parameters.timelimit.set(200) # Use max 200 seconds for solving
        #problem.parameters.mip.limits.populate.set(20) # Find max 20 solutions (=default)
   
    else:
        logger.warning("You are trying to run 'Hamlet Hot Rig' with %s. This might not end well." %
                      model.solver.interface.__name__.split(".")[-1])
    
    # Remove reactions that are blocked: no flux through these reactions possible. This will reduce the search space for the solver, if not done already.
    if remove_blocked:
        blocked_reactions = cameo.flux_analysis.analysis.find_blocked_reactions(fused_model)
        fused_model.remove_reactions(blocked_reactions)
        logger.debug("Removed " + str(len(blocked_reactions)) + " reactions that were blocked")
    
    # Make dual
    model_without_target = model.copy() # This variable looks unnecessary, but is kept out of fear of messing stuff up
    model_without_target.optimize()
    dual_problem = convert_linear_problem_to_dual(model_without_target.solver)
    logger.debug("Dual problem successfully created")

    # Combine primal and dual
    primal_problem = model.solver

    for var in dual_problem.variables:  # All variables in the dual are copied to the primal
        var = primal_problem.interface.Variable.clone(var)
        primal_problem.add(var)
    for const in dual_problem.constraints:  # All constraints in the dual are copied to the primal
        const = primal_problem.interface.Constraint.clone(const, model=primal_problem)
        primal_problem.add(const)
    logger.debug("Dual and primal combined")

    dual_problem.optimize()
    
    
    # Dictionaries to hold the binary control variables: 
    native_y_vars = {}       # 1 for knockout, 0 for active
    heterologous_y_vars = {} # 1 for knockin, 0 for inactive
    medium_y_vars = {}       # 1 for medium addition (up to -10), 0 for no addition
    
    # Now the fun stuff
    constrained_dual_vars = set()
    
    # Fill the dictionaries with binary variables
    # For the knockouts: only for the native reactions which are not exchanges
    for reaction in [r for r in model.reactions - excluded_reactions if r.type == "native"]:

        # Add constraint variables
        interface = model.solver.interface
        y_var = interface.Variable("y_" + reaction.id, type="binary")

        # Constrain the primal: flux through reactions maximum within (-1000, 1000), or smaller boundaries defined before
        model.solver.add(interface.Constraint(reaction.flux_expression - 1000 * (1 - y_var), ub=0, name="primal_y_const_"+reaction.id+"_ub"))
        model.solver.add(interface.Constraint(reaction.flux_expression + 1000 * (1 - y_var), lb=0, name="primal_y_const_"+reaction.id+"_lb")) 

        # Constrain the dual
        constrained_vars = []

        if reaction.upper_bound != 0:
            dual_forward_ub = model.solver.variables["dual_" + reaction.forward_variable.name + "_ub"]
            model.solver.add(interface.Constraint(dual_forward_ub - 1000 * y_var, ub=0))
            constrained_vars.append(dual_forward_ub)
        if reaction.lower_bound != 0:
            dual_reverse_ub = model.solver.variables["dual_" + reaction.reverse_variable.name + "_ub"] 
            model.solver.add(interface.Constraint(dual_reverse_ub - 1000 * y_var, ub=0))
            constrained_vars.append(dual_reverse_ub)
        constrained_dual_vars.update(constrained_vars)

        # Add to dictionary with binary variables for native reactions:
        native_y_vars[y_var] = reaction


    # For the knockins and medium additions:
    for reaction in [r for r in model.reactions - excluded_reactions if r.type == "heterologous" or r.type == "medium"]:          
        # Add constraint variables
        interface = model.solver.interface
        y_var = interface.Variable("y_" + reaction.id, type="binary")

        # Constrain the primal: flux through reactions maximum within (-1000, 1000), or smaller boundaries defined before
        model.solver.add(interface.Constraint(reaction.flux_expression - 1000 * y_var, ub=0, name="primal_y_const_"+reaction.id+"_ub"))
        model.solver.add(interface.Constraint(reaction.flux_expression + 1000 * y_var, lb=0, name="primal_y_const_"+reaction.id+"_lb"))  

        # Constrain the dual
        constrained_vars = []

        if reaction.upper_bound != 0:
            dual_forward_ub = model.solver.variables["dual_" + reaction.forward_variable.name + "_ub"]
            model.solver.add(interface.Constraint(dual_forward_ub - 1000 * (1 - y_var), ub=0))
            constrained_vars.append(dual_forward_ub)
        if reaction.lower_bound != 0:
            dual_reverse_ub = model.solver.variables["dual_" + reaction.reverse_variable.name + "_ub"]
            model.solver.add(interface.Constraint(dual_reverse_ub - 1000 * (1 - y_var), ub=0))
            constrained_vars.append(dual_reverse_ub)
        constrained_dual_vars.update(constrained_vars)

        # Add y variable to the corresponding modifications dictionary
        if reaction.type == "medium":
            medium_y_vars[y_var] = reaction         
        elif reaction.type == "heterologous":
            heterologous_y_vars[y_var] = reaction
            
    logger.debug("Control variables created")
    

    # Set the objective
    primal_objective = model.solver.objective
    dual_objective = interface.Objective.clone(
        dual_problem.objective, model=model.solver
    )

    reduced_expression = optlang.symbolics.Add(
        *((c * v) for v, c in dual_objective.expression.as_coefficients_dict().items()
        if v not in constrained_dual_vars)
    )
    dual_objective = interface.Objective(reduced_expression, direction=dual_objective.direction)

    full_objective = interface.Objective(primal_objective.expression - dual_objective.expression, direction="max")
    model.objective = full_objective
    logger.debug("Objective created")
    
        
    # Add number of knockouts constraint. ub=K+1 as the target will be forced to be 1 as well
    knockout_number_constraint = model.solver.interface.Constraint(
        optlang.symbolics.Add(*native_y_vars), lb=0, ub=K+1, name="number_of_knockouts_constraint"
    )
    model.solver.add(knockout_number_constraint)
   
    # Add number of knockins constraint
    knockin_number_constraint = model.solver.interface.Constraint(
        optlang.symbolics.Add(*heterologous_y_vars), lb=0, ub=L, name="number_of_knockins_constraint"
    )
    model.solver.add(knockin_number_constraint)
  
    # Add number of medium additions constraint
    medium_additions_number_constraint = model.solver.interface.Constraint(
        optlang.symbolics.Add(*medium_y_vars), lb=0, ub=M, name="number_of_medium_additions_constraint"
    )
    model.solver.add(medium_additions_number_constraint)
    
    logger.debug("Added constraint for number of knockouts, knockins and medium additions")

    
    # Force target control variable to be 1, but allow flux in primal
    model.solver.constraints["primal_y_const_" + target + "_lb"].lb = -2000
    model.solver.constraints["primal_y_const_" + target + "_ub"].ub = 2000
    model.variables["y_" + target].lb = 1
    
    return model, native_y_vars, heterologous_y_vars, medium_y_vars

# Create the MILP from the fused model, with at most K knockouts, L knockins and M medium additions allowed.    
milp, native_y_vars, heterologous_y_vars, medium_y_vars = setup_milp(fused_model, target_id,\
                    medium_metabolites, K=number_of_knockouts, L=number_of_knockins, M=number_of_medium_additions, remove_blocked=False, exclude_reaction_ids=exclude_reaction_ids)
                    
print("# Candidates for knock-outs:", len(native_y_vars))
print("# Candidates for knock-ins:", len(heterologous_y_vars))
print("# Candidates for medium additions:", len(medium_y_vars))

# Create a runname that is used for the filenames belonging to this run
def create_runname(milp):
    """
    Create a runname for the current run:
    YYYYMMDD_HHMM__<model name>__<target ID>
    """
    import datetime
    # Retrieve current time
    date_time = datetime.datetime.now().strftime("%Y%m%d_%H%M")
    # Retrieve target from model, to make sure that this is the correct target 
    # (in contrast to an extra input to this function than can be forgotten to be changed)
    target = [v.VarName[2:] for v in milp.solver.problem.getVars() if v.LB == 1 and v.VarName.startswith("y_")][0]
    runname = "__".join([date_time, milp.id, target])
    return runname

runname = create_runname(milp)

# Set parameters for solving with the Gurobi solution pool:
milp.solver.problem.params.PoolSearchMode = 2 #0 for only finding the optimum, 1 for finding more solutions (but no quality guaranteed), 2 for finding the n best possible solutions 
milp.solver.problem.params.PoolSolutions = 2000000 # Number of solutions kept when finding the optimal solution, max 2,000,000,000
milp.solver.problem.params.PoolGap = 0.99 # only store solutions within 90% of the optimal objective value
milp.solver.problem.params.TimeLimit = 130000 # Set a time limit used for finding the solution.

# Solve the milp: create a solution pool. Time how long the solver took to optimize
from timeit import default_timer as timer
start = timer()
milp.optimize()
end = timer()
print("Optimization time:", end - start, "s /", (end - start)/60, "m /", (end - start)/3600, "h")
optimization_time_sec = end - start

# Write all parameters used to a file for later reference
def write_parameters_to_file(milp, runname, excluded_reaction_ids, optimization_time_sec):
    filename = "%s__details.txt" % runname
    split_runname = runname.split("__")
    date_time = split_runname[0][:4] + "-" + split_runname[0][4:6] + "-" + split_runname[0][6:8] + " " + split_runname[0][9:11] + ":" + split_runname[0][11:13]
    with open(filename, "w") as output_file:
        output_file.write("Run details for %s:\n" % runname)
        output_file.write("Date: %s\n" % date_time)
        output_file.write("Model: %s\n" % split_runname[1])
        output_file.write("Target: %s\n" % split_runname[2])

        # In case no knockins or medium additions might be allowed:
        ki_var = milp.solver.problem.getVarByName("number_of_knockins_constraint_aux")
        ki_number = int(ki_var.x) if ki_var is not None else 0
        ma_var = milp.solver.problem.getVarByName("number_of_medium_additions_constraint_aux")
        ma_number = int(ma_var.x) if ma_var is not None else 0    
            
        output_file.write("Knockouts allowed: %i\n" % (int(milp.solver.problem.getVarByName("number_of_knockouts_constraint_aux").x)-1))
        output_file.write("Knockins allowed: %i\n" % ki_number)
        output_file.write("Medium additions allowed: %i\n" % ma_number)
        output_file.write("Number of solutions found: %s\n" % milp.solver.problem.solcount)
        solver_settings, solver_parameters = str(milp.solver.problem)[1:-1].split(", Parameter changes: ")
        solver_settings = solver_settings.replace(" Unnamed: ", "\n\t")
        output_file.write("\nSolver settings: \n\t%s\n" % solver_settings)
        output_file.write("Optimization time: %0.0f s, %0.1f m, %0.2f h\n" % (optimization_time_sec, optimization_time_sec/60, optimization_time_sec/3600))
        output_file.write("Solver status: %s\n" % milp.solver.status)
        solver_parameters = solver_parameters.split(", ")
        output_file.write("Parameter changes:\n")
        [output_file.write("\t%s\n" % setting) for setting in solver_parameters]
        output_file.write("\nReactions excluded from modification:\n")
        output_file.write(", ".join(excluded_reaction_ids))
        output_file.write("\n")
                          
write_parameters_to_file(milp, runname, exclude_reaction_ids, optimization_time_sec)

# Extract the solutions from the solution pool and put them in a list of dicts
def extract_solutions_from_gurobi_pool(milp, target_id, biomass_id):
    """
    Extracts the individual solutions from a MILP solution pool created by the Gurobi solver.
    Returns the solution pool and a header with information to accompany the list of dicts.
    """
    if "gurobi_interface" not in milp.solver.interface.__name__:
        raise ValueError("The solver for the MILP is not Gurobi. This function only works for Gurobi created solution pools.")
    if target_id not in milp.solver.problem.getVars():
        raise ValueError("Specified target reaction ID %s not found in the variable list." % target_id)
    if biomass_id not in milp.solver.problem.getVars():
        raise ValueError("Specified biomass reaction ID %s not found in the variable list." % biomass_id)
        
    solution_pool = list()
    for solution_number in range(milp.solver.problem.solcount):
        solution = dict()
        milp.solver.problem.params.SolutionNumber = solution_number
        solution["number"] = solution_number
        # Note that the objective is the growth coupling potential, not the growth or production rate
        solution["objective"] = milp.solver.problem.PoolObjVal 
        
        # Prepare dict for looping over variables
        # Assigning all keys here to put them in a certain order in the dict
        target_flux_name = "target_flux_(%s)" % target_id
        biomass_id_reverse = biomass_id + "_reverse_"
        target_id_reverse = target_id + "_reverse_"
        
        solution["growth_coupling_slope"] = None
        solution[target_flux_name] = 0
        solution["growth_rate"] = 0
        solution["growth_without_target"] = None
        #solution["target/biomass"] = None
        solution["knockout"] = list()
        solution["knockin"] = list()
        solution["medium_addition"] = list()

        
        for v in milp.solver.problem.getVars():
            if v.VarName.startswith("y_") and v.xn > 0.99:
                reaction_id = v.VarName[2:] #removing "y_"
                reaction_type = milp.reactions.get_by_id(reaction_id).type
                if reaction_type == "native" and reaction_id != target_id:
                    solution["knockout"].append(reaction_id)
                elif reaction_type == "heterologous":
                    solution["knockin"].append(reaction_id)
                elif reaction_type == "medium":
                    solution["medium_addition"].append(reaction_id)
            
            # Extract target flux and growth rate. Subtract the forward and reverse fluxes to get the nett flux.
            if v.VarName == target_id:
                solution[target_flux_name] += v.xn
                # In case target_id or biomass_id are not found, and these values are assigned None to start with:
                #solution["target_flux"] = v.xn if solution["target_flux"] is None else solution["target_flux"] + v.xn
            if v.VarName.startswith(target_id_reverse):
                solution[target_flux_name] -= v.xn
            if v.VarName == biomass_id:
                solution["growth_rate"] += v.xn
            if v.VarName.startswith(biomass_id_reverse):
                solution["growth_rate"] -= v.xn
        
        #if round(solution["growth_rate"],7) > 0:
        #    solution["target/biomass"] = solution[target_flux_name]/solution["growth_rate"]
            
        solution["growth_without_target"] = solution["growth_rate"] - solution["objective"]
        if round(solution["objective"],7) > 0:
            solution["growth_coupling_slope"] = solution[target_flux_name] / solution["objective"]
            
        solution_pool.append(solution)
        
        
    return solution_pool

solution_pool= extract_solutions_from_gurobi_pool(milp, target_id, biomass_id)

# Write the list of dicts with all solutions to json and csv files
def write_solution_pool_to_json(solution_pool, runname, suffix="", indent=4):
    """Write the list of dicts to json format for easy retrieval of the data structure"""
    import json
    filename = "%s__solution_pool%s.json" % (runname, suffix)
    with open(filename, "w") as output_file:
        json.dump(solution_pool, output_file, indent=indent)
        
write_solution_pool_to_json(solution_pool, runname, suffix="_all")

def write_solution_pool_to_csv(solution_pool, runname, suffix=""):
    import unicodecsv as csv
    filename = "%s__solution_pool%s.csv" % (runname, suffix)
    keys = solution_pool[0].keys()
    with open(filename, 'wb') as output_file:
        dict_writer = csv.DictWriter(output_file, keys)
        dict_writer.writeheader()
        dict_writer.writerows(solution_pool)
        
write_solution_pool_to_csv(solution_pool, runname, suffix="_all")

# Create a list of solutions that only contains the minimal modifications needed for the growth coupling
# Thus not the modifications that simply enhance growth rate.
from itertools import chain, combinations
def powerset(iterable):
    """
    From the itertools docs:
    Function that creates all possible combination sets out of a list of elements.
    i.e. powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))
    
def select_minimal_designs(native_model, milp, solution_pool, target_id=None, biomass_id=None):
    """
    Selects the minimal set of modifications from the solution pool, 
    e.g. filters out solutions that have additional modifications that do not contribute to the growth coupling
    """
    
    #Autodetect target and biomass IDs if not given:
    if target_id is None:
        target_id = [v.VarName[2:] for v in milp.solver.problem.getVars() if v.LB == 1 and v.VarName.startswith("y_")][0]
    if biomass_id is None:
        biomass_id = [r.id for r in model.reactions if "biomass" in r.id.lower()][0]
    
    target_flux_name = "target_flux_(%s)" % target_id
    minimal_solution_pool = list()
    minimal_designs_so_far = list()
    
    # Loop over all solutions in the solution pool
    for solution in solution_pool:
        ko = solution["knockout"]
        ki = solution["knockin"]
        ma = solution["medium_addition"]      
        new_reactions = ki + ma
        modifications = ko + ki + ma
        
        already_found = False
        for design in minimal_designs_so_far:
            if all(reaction in modifications for reaction in design):
                already_found = True
                break
        
        if not already_found:
            design_found = False
            minimal_combination = dict()
            with native_model as model:
                # Add the extra reactions in this solution to the native model
                for r in new_reactions:
                    model.add_reaction(milp.reactions.get_by_id(r).copy())
                # Loop over all combinations that these genes can be allowed to have flux
                # starting with the individual genes, then pairs, etc.
                for combination in list(powerset(modifications)):
                    with model as test_model:
                        for reaction in ko: 
                            if reaction in combination:
                                test_model.reactions.get_by_id(reaction).knock_out()
                        for reaction in new_reactions: 
                            if reaction not in combination:
                                test_model.reactions.get_by_id(reaction).knock_out()
                        test_model.optimize()
                        target_flux = test_model.reactions.get_by_id(target_id).flux
                        biomass_flux_max = test_model.reactions.get_by_id(biomass_id).flux
                        test_model.reactions.get_by_id(target_id).bounds = (0, 0)
                        biomass_flux_min = test_model.slim_optimize()
                        biomass_flux_delta = biomass_flux_max - biomass_flux_min
                        if round(biomass_flux_delta,7) != 0 and round(target_flux/biomass_flux_delta,5) == round(solution["growth_coupling_slope"],5):
                            minimal_combination["objective"] = round(biomass_flux_delta,7)
                            minimal_combination["growth_coupling_slope"] = round(target_flux/biomass_flux_delta,7)
                            minimal_combination[target_flux_name] = round(target_flux,7)
                            minimal_combination["growth_rate"] = round(biomass_flux_max,7)
                            minimal_combination["growth_without_target"] = round(biomass_flux_min,7)
                            minimal_combination["knockout"] = [r for r in ko if r in combination]
                            minimal_combination["knockin"] = [r for r in ki if r in combination]
                            minimal_combination["medium_addition"] = [r for r in ma if r in combination]
                            design_found = True
                            break

            if design_found:
                # Only keep minimal solutions that are unique
                if minimal_solution_pool == list():
                    minimal_solution_pool.append(minimal_combination)
                    minimal_designs_so_far.append(minimal_combination["knockout"]+minimal_combination["knockin"]+minimal_combination["medium_addition"])

                else:
                    for key in minimal_combination.keys():
                        if minimal_combination[key] not in [value for i in range(len(minimal_solution_pool)) \
                                                      for value in minimal_solution_pool[i].values()]:
                            minimal_solution_pool.append(minimal_combination)
                            minimal_designs_so_far.append(minimal_combination["knockout"]+minimal_combination["knockin"]+minimal_combination["medium_addition"])
                            break
            else:
                print("Warning: No minimal design found for: %s" % ", ".join(modifications))
            
    return minimal_solution_pool

minimal_solution_pool = select_minimal_designs(model, milp, solution_pool, target_id, biomass_id)
write_solution_pool_to_json(minimal_solution_pool, runname, suffix="_minimal")
write_solution_pool_to_csv(minimal_solution_pool, runname, suffix="_minimal")

# Make a file with the full reactions and metabolite names involved in the modifications
# to get a bit of an overview what these designs entail.
def write_solution_pool_to_verbose_txt(solution_pool, milp, runname, suffix):
    target_id = [v.VarName[2:] for v in milp.solver.problem.getVars() if v.LB == 1 and v.VarName.startswith("y_")][0]
    
    with open("%s__solution_pool%s.txt" % (runname, suffix), "w") as output_file:
        output_file.write("Solution pool for run %s:\n" % runname)
        for i, solution in enumerate(solution_pool):
            output_file.write("\n\n>>> Solution %i <<<\n" % i)
            output_file.write("Target flux: %f    Growth rate: %f\n" % (solution["target_flux_(%s)" % target_id], solution["growth_rate"]))
            output_file.write("Growth-coupling slope: %f    Growth without target: %f\n" % (solution["growth_coupling_slope"], solution["growth_without_target"]))
            if solution["knockout"] != []:
                output_file.write("\n> Knockout:\n")
                for r_id in solution["knockout"]:
                    ko = milp.reactions.get_by_id(r_id)
                    output_file.write("%s (%s):\n    %s   %s\n" % (r_id, ko.name, ko.reaction, ko.bounds))
                    for m in ko.reactants + ko.products:
                        output_file.write("        %s: %s\n" % (m.id, m.name))
            if solution["knockin"] != []:
                output_file.write("\n> Knockin:\n")
                for r_id in solution["knockin"]:
                    ki = milp.reactions.get_by_id(r_id)
                    EC_string = " EC: %s" % ki.annotation.get("EC").replace(";", "; ") if ki.annotation.get("EC") is not None else ""
                    output_file.write("%s (%s, %s):\n    %s   %s\n" % (r_id, ki.name, EC_string, ki.reaction, ki.bounds))
                    for m in ki.reactants + ki.products:
                        output_file.write("        %s: %s\n" % (m.id, m.name))
            if solution["medium_addition"] != []:
                output_file.write("\n> Add to the medium:\n")
                for r_id in solution["medium_addition"]:
                    ma = milp.metabolites.get_by_id(r_id[4:])
                    output_file.write("%s: %s (%s)\n" % (r_id, ma.name, ma.formula))
                
write_solution_pool_to_verbose_txt(minimal_solution_pool, milp, runname, "_minimal")

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.patches import Polygon
from matplotlib.ticker import AutoMinorLocator

# Create production envelopes for all solutions in the minimal solution pool
# to visualize the extent of growth coupling.
def create_production_envelope(orig_model, target_id, biomass_id, plottitle="Production envelope", filename="productionenvelope.png", xmax=None, ymax=None, points=100):
    """Creates a production envelope using matplotlib pyplot's Polygon."""
    with orig_model as model:
        model.objective = biomass_id
        model.objective_direction = "max"
        max_growth_rate = model.slim_optimize()
        growth_rate_interval = [max_growth_rate/(points-1) * i for i in range(0,points)]
        model.objective = target_id
        max_target_flux = []
        min_target_flux = []
        for growth_rate in growth_rate_interval:
            model.reactions.get_by_id(biomass_id).bounds = (growth_rate, growth_rate)
            model.objective_direction = "max"
            max_target_flux.append(model.slim_optimize())
            model.objective_direction = "min"
            min_target_flux.append(model.slim_optimize())        
    # Make one list of tuples out of the minimal and maximal values
    max_target_flux.extend(min_target_flux[::-1])
    growth_rate_interval.extend(growth_rate_interval[::-1])
    ppp = list(zip(growth_rate_interval, max_target_flux))
    # Make into numpy array
    dt=np.dtype('float,float')
    np_ppp = np.array(ppp,dtype=dt)
    # Plot with matplotlip pyplot
    fig = plt.figure(10, figsize=(5,3.5), dpi=300)
    ax = fig.add_subplot(111)
    patch = Polygon(ppp, edgecolor="#fc9402" , facecolor="#ffe3b7", linewidth=2, closed=True)
    ax.add_patch(patch)
    if xmax is None:
        xmax = max_growth_rate * 1.14
    if ymax is None:
        ymax = max(max_target_flux) * 1.2
    xrange = [0, xmax]
    yrange = [0, ymax]
    ax.set_xlim(*xrange)
    ax.set_ylim(*yrange)
    plt.xlabel(r'%s (h$^{-1}$)'% biomass_id, fontsize=8)
    plt.ylabel(r'%s (mmol gDW$^{-1}$ h$^{-1}$)'% target_id, fontsize=8)
    ax.set_title(" ")
    plt.grid(b=True, which='major', axis='both', alpha=0.2)
    minorLocator = AutoMinorLocator()
    ax.xaxis.set_minor_locator(minorLocator)
    minorLocator = AutoMinorLocator()
    ax.yaxis.set_minor_locator(minorLocator)
    ax.text(0, 1.02, "%s" % plottitle, fontsize=9,
        horizontalalignment='left',
        verticalalignment='bottom',
        transform=ax.transAxes)
    plt.tight_layout()
    plt.savefig("%s" % filename)
    plt.close(fig)
    

def create_solution_pool_production_envelopes(native_model, milp, solution_pool, target_id, biomass_id, runname):
    """
    Creates phenotypic phase planes for all solutions in the solution pool.
    Stores them as png in a separate folder.
    """
    # Create a folder to put all envelopes in
    dirname = "%s_prod_envs" % runname
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    
    # Make a production envelope of the native model to start with
    filename = "%s/%s__%s__no_modifications" % (dirname, native_model.id, target_id)
    xmax = native_model.slim_optimize() * 1.14
    create_production_envelope(native_model, target_id, biomass_id, plottitle="%s with %s target" % (native_model.id, target_id), filename=filename, xmax=xmax)

    # Loop over all solution, modify the model and make a production envelope
    for i, solution in enumerate(solution_pool):
        with native_model as model:
            for ko in solution["knockout"]:
                model.reactions.get_by_id(ko).knock_out()
            for ki in solution["knockin"] + solution["medium_addition"]:
                new = milp.reactions.get_by_id(ki).copy()
                model.add_reaction(new)
            filename = "%s/%s__%s__solution_%i" % (dirname, model.id, target_id, i)
            plottitle = "%s with " % model.id
            if solution["knockout"] != []:
                plottitle += "KO: %s; " % ", ".join(solution["knockout"])
            if solution["knockin"] != []:
                plottitle += "KI: %s; " % ", ".join(solution["knockin"])    
            if solution["medium_addition"] != []:
                plottitle += "ADD: %s; " % ", ".join([r_id[4:] for r_id in solution["medium_addition"]])
            
            create_production_envelope(model, target_id, biomass_id, plottitle=plottitle, filename=filename, xmax=xmax)

        
create_solution_pool_production_envelopes(model, milp, minimal_solution_pool, target_id, biomass_id, runname)

print("Done. Have fun analysing the data.")
print('  (\\_/)\n =(^.^)=\n  (   )\n (")_(")')
